(ns puppetlabs.rbac.http.middleware
  (:require [clojure.string :as str]
            [clojure.tools.logging :as log]
            [cheshire.core :as json]
            [clj-stacktrace.repl]
            [slingshot.slingshot :refer [try+]]
            [ring.util.codec :refer [url-encode]]
            [ring.util.response :refer [redirect response status header]]
            [puppetlabs.schema-tools :refer [explain-and-simplify-exception-data]]))

(defn json?
  "Given a ring request map, checks the Content-Type and Accept header
  to determine if this request desires json back"
  [request]
  (let [header-match #(re-find #"(?i)^application/json.*" (get-in request [:headers %] ""))]
    (some header-match ["accept" "content-type" "Accept" "Content-Type"])))

(defn build-response
  "Given a ring request map, an HTTP status, a string body, and an
  optional map of additional headers, return a response map suitable
  for ring."
  [request status-code error-map]
  (let [resp (if (json? request)
               (response (json/encode error-map))
               (response (:msg error-map)))]
    (-> resp
        (status status-code)
        (header "Content-Type" (if (json? request) "application/json" "text/html")))))

(defn handle-unauthenticated
  "Given a ring request map and a (possibly empty) redirect string,
  generate either a 401 (for json requests) or a 302."
  [login-path request redirect-url]
  (if (json? request)
    (build-response request 401 "Route requires authentication.")
    (let [redirect-url (if (empty? redirect-url) "" (format "?redirect=%s" (url-encode redirect-url)))]
      (redirect (str login-path redirect-url)))))

(defn- log-exception
  "Log an exception including request method and URI"
  [request exception]
  (let [method (-> request :request-method name str/upper-case)
        message (str method " " (:uri request))]
    (log/warn message (clj-stacktrace.repl/pst-str exception))))

(defn wrap-id-filter-errors
  "Middlware handling validation error when a url contains an invalid filter
  on id (/users?id=%^&#$%)"
  [handler]
  (fn [request]
    (try+ (handler request)
      (catch [:kind :puppetlabs.rbac/invalid-id-filter] _
        (build-response request 400
                        {:kind "invalid-id-filter"
                         :msg "Invalid ID filter list"
                         :details {}})))))

(defn wrap-id-validation-errors
  "Middleware handling validation error when a url contains an id that doesn't
  match the id in submitted data"
  [handler]
  (fn [request]
    (try+ (handler request)
      (catch [:kind :puppetlabs.rbac/inconsistent-id]
        {:keys [url-id body-id]}
        (build-response request 400
                        {:kind "inconsistent-id"
                         :msg "The ID in the url does not match the ID in the submitted data."
                         :details {:url-id url-id
                                   :body-id body-id}})))))

(defn wrap-schema-errors
  "Middleware to handle schema validation failures"
  [handler]
  (fn [request]
    (try+ (handler request)
      (catch [:kind :puppetlabs.rbac/schema-violation] exc-data
        (let [{:keys [schema value error]} (explain-and-simplify-exception-data exc-data)]
          (build-response request 400
                          {:kind "schema-violation"
                           :msg (str "The object(s) submitted did not conform to the schema."
                                     " The problem is: " (seq error))
                           :details {:submitted value
                                     :schema schema
                                     :error error}}))))))

(defn wrap-authn-errors
  "Middlware to handle authentication errors"
  [handler login-path]
  (fn [request]
    (try+ (handler request)
      (catch [:kind :puppetlabs.rbac/user-unauthenticated] {:keys [redirect]}
        (handle-unauthenticated login-path request redirect))
      (catch [:kind :puppetlabs.rbac/user-revoked] s
        (build-response request 401 {:msg "This account has been revoked."})))))

(defn authz-error-response
  "Returns a 403 forbidden error with `msg` and `kind`"
  [request {:keys [kind msg]}]
  (build-response request 403 {:kind (str (name kind))
                               :msg msg}))

(defn wrap-authz-errors
  "Middleware for handling issues related to authz"
  [handler]
  (fn [request]
    (try+ (handler request)
          (catch [:kind :puppetlabs.rbac/permission-denied] error-map
            (authz-error-response request error-map)))))

(defn wrap-data-update-errors
  "Middleware for handling issues related to data updates"
  [handler]
  (fn [request]
    (try+ (handler request)
          (catch [:kind :puppetlabs.rbac.storage/admin-user-immutable] error-map
            (authz-error-response request error-map))
          (catch [:kind :puppetlabs.rbac.storage/admin-user-not-in-admin-role] error-map
            (authz-error-response request error-map))
          (catch [:kind :puppetlabs.rbac.storage/default-roles-immutable] error-map
            (authz-error-response request error-map)))))

(defn wrap-catchall-500
  "Middleware to handle any unexpected errors by logging the full stacktrace
  and returning a 500 Server Error"
  [handler]
  (fn [request]
    (try (handler request)
      (catch Throwable e
        (let [root-e (if (instance? Iterable e)
                       (-> e seq last)
                       e)]
          (log-exception request e)
          (build-response request 500 {:msg (format "Server error: %s" (.getMessage root-e))}))))))

(defn wrap-errors
  "This middleware composes all the error handling middleware in this namespace."
  [handler login-path]
  (-> handler
      wrap-id-filter-errors
      wrap-id-validation-errors
      wrap-schema-errors
      (wrap-authn-errors login-path)
      wrap-data-update-errors
      wrap-authz-errors
      wrap-catchall-500))
